from __future__ import annotations

from typing import Callable

import pygame
from pygame import rect

try:
    from typing import Self
except ImportError:  # pragma: no cover - Python 3.10 fallback
    from typing_extensions import Self

from ..entities_constants import (
    INTERNAL_WALL_BEVEL_DEPTH,
    INTERNAL_WALL_HEALTH,
    STEEL_BEAM_HEALTH,
)
from ..render_assets import (
    EnvironmentPalette,
    build_beveled_polygon,
    build_rubble_wall_surface,
    paint_steel_beam_surface,
    paint_wall_damage_overlay,
    paint_wall_surface,
    resolve_steel_beam_colors,
    resolve_wall_colors,
    resolve_wall_outline_color,
    rubble_offset_for_size,
    RUBBLE_ROTATION_DEG,
)
from .movement import (
    _circle_polygon_collision,
    _circle_rect_collision,
    _rect_polygon_collision,
)

_WALL_INDEX_DIRTY = False
_WALL_DAMAGE_OVERLAY_SEED = 1337


def _draw_dotted_rect_outline(
    image: pygame.Surface,
    *,
    color: tuple[int, int, int],
    step: int = 3,
) -> None:
    """Draw a 1px dotted outline around the whole surface rectangle."""
    width, height = image.get_size()
    if width <= 1 or height <= 1:
        return
    dot_step = max(2, int(step))
    max_x = width - 1
    max_y = height - 1
    for x in range(0, width, dot_step):
        image.set_at((x, 0), color)
        image.set_at((x, max_y), color)
    for y in range(0, height, dot_step):
        image.set_at((0, y), color)
        image.set_at((max_x, y), color)
    image.set_at((0, 0), color)
    image.set_at((max_x, 0), color)
    image.set_at((0, max_y), color)
    image.set_at((max_x, max_y), color)


def _damage_overlay_variant_index(
    *,
    x: int,
    y: int,
    width: int,
    height: int,
) -> int:
    cell_w = max(1, int(width))
    cell_h = max(1, int(height))
    cell_x = int(x) // cell_w
    cell_y = int(y) // cell_h
    return (cell_x % 3) + ((cell_y % 3) * 3)


def _mark_wall_index_dirty() -> None:
    global _WALL_INDEX_DIRTY
    _WALL_INDEX_DIRTY = True


def consume_wall_index_dirty() -> bool:
    global _WALL_INDEX_DIRTY
    if not _WALL_INDEX_DIRTY:
        return False
    _WALL_INDEX_DIRTY = False
    return True


class Wall(pygame.sprite.Sprite):
    def __init__(
        self: Self,
        x: int,
        y: int,
        width: int,
        height: int,
        *,
        health: int = INTERNAL_WALL_HEALTH,
        palette: EnvironmentPalette | None = None,
        palette_category: str = "inner_wall",
        bevel_depth: int = INTERNAL_WALL_BEVEL_DEPTH,
        bevel_mask: tuple[bool, bool, bool, bool] | None = None,
        draw_bottom_side: bool = False,
        bottom_side_ratio: float = 0.1,
        side_shade_ratio: float = 0.9,
        damage_overlay_seed: int = _WALL_DAMAGE_OVERLAY_SEED,
        on_destroy: Callable[[Self], None] | None = None,
    ) -> None:
        super().__init__()
        safe_width = max(1, width)
        safe_height = max(1, height)
        self.image = pygame.Surface((safe_width, safe_height), pygame.SRCALPHA)
        self.palette = palette
        self.palette_category = palette_category
        self.health = health
        self.max_health = max(1, health)
        self.on_destroy = on_destroy
        self.bevel_depth = max(0, bevel_depth)
        self.bevel_mask = bevel_mask or (False, False, False, False)
        self.draw_bottom_side = draw_bottom_side
        self.bottom_side_ratio = max(0.0, bottom_side_ratio)
        self.side_shade_ratio = max(0.0, min(1.0, side_shade_ratio))
        self._damage_visual_seed = damage_overlay_seed
        self._damage_overlay_variant = _damage_overlay_variant_index(
            x=x,
            y=y,
            width=safe_width,
            height=safe_height,
        )
        self._local_polygon = _build_beveled_polygon(
            safe_width, safe_height, self.bevel_depth, self.bevel_mask
        )
        self._update_color()
        self.rect = self.image.get_rect(topleft=(x, y))
        # Keep collision rectangular even when beveled visually.
        self._collision_polygon = None

    def _take_damage(self: Self, *, amount: int = 1) -> None:
        if self.health > 0:
            self.health -= amount
            self._update_color()
            if self.health <= 0:
                if self.on_destroy:
                    try:
                        self.on_destroy(self)
                    except Exception as exc:
                        print(f"Wall destroy callback failed: {exc}")
                _mark_wall_index_dirty()
                self.kill()

    def _update_color(self: Self) -> None:
        if self.health <= 0:
            health_ratio = 0.0
        else:
            health_ratio = max(0.0, self.health / self.max_health)
        fill_color, _border_color = resolve_wall_colors(
            health_ratio=health_ratio,
            palette_category=self.palette_category,
            palette=self.palette,
        )
        outline_color = resolve_wall_outline_color(
            palette_category=self.palette_category,
            palette=self.palette,
        )
        use_dotted_outline = self.palette_category == "inner_wall" and 0.0 < health_ratio < 1.0
        paint_wall_surface(
            self.image,
            fill_color=fill_color,
            outline_color=outline_color,
            bevel_depth=self.bevel_depth,
            bevel_mask=self.bevel_mask,
            draw_bottom_side=self.draw_bottom_side,
            bottom_side_ratio=self.bottom_side_ratio,
            side_shade_ratio=self.side_shade_ratio,
            draw_outline=not use_dotted_outline,
        )
        self._paint_panel_relief(fill_color=fill_color)
        self._paint_damage_marks(health_ratio=health_ratio)
        self._apply_damage_alpha(health_ratio=health_ratio)

    def _paint_panel_relief(
        self: Self,
        *,
        fill_color: tuple[int, int, int],
    ) -> None:
        w, h = self.image.get_size()
        if w <= 2 or h <= 2:
            return
        # Keep panel placement stable across tiles regardless of whether their
        # bottom side is currently visible; this avoids per-tile vertical drift.
        reference_side_height = max(1, int(h * max(0.0, self.bottom_side_ratio)))
        top_height = max(1, h - reference_side_height)
        top_rect = pygame.Rect(0, 0, w, top_height)

        # Recessed panel effect inside the top face.
        inset = max(2, min(w, top_height) // 6)
        # Define the panel base rectangle first, independent of bevel details.
        panel_rect = pygame.Rect(
            top_rect.left + inset,
            top_rect.top + inset,
            max(1, top_rect.width - inset * 2),
            max(1, top_rect.height - inset * 2),
        )
        # Make the decoration square slightly smaller on all sides (+1px inset each side).
        panel_rect = panel_rect.inflate(-2, -2)
        if panel_rect.width < 6 or panel_rect.height < 6:
            return
        panel_color = (
            max(0, int(fill_color[0] * 0.96)),
            max(0, int(fill_color[1] * 0.96)),
            max(0, int(fill_color[2] * 0.96)),
        )
        panel_border = (
            max(0, int(fill_color[0] * 0.72)),
            max(0, int(fill_color[1] * 0.72)),
            max(0, int(fill_color[2] * 0.72)),
        )
        panel_overlay = pygame.Surface(panel_rect.size, pygame.SRCALPHA)
        local_rect = panel_overlay.get_rect()
        pygame.draw.rect(panel_overlay, panel_color, local_rect)
        pygame.draw.rect(panel_overlay, panel_border, local_rect, width=1)
        # North edge accent: draw as an inset 2px band to avoid corner misalignment.
        top_band = pygame.Rect(
            1,
            1,
            max(1, local_rect.width - 2),
            max(1, min(2, local_rect.height - 2)),
        )
        if top_band.width > 0 and top_band.height > 0:
            pygame.draw.rect(panel_overlay, panel_border, top_band)
        # Trim 1px from each corner for a subtle rounded-corner read.
        corner_points = [
            (0, 0),
            (local_rect.width - 1, 0),
            (0, local_rect.height - 1),
            (local_rect.width - 1, local_rect.height - 1),
        ]
        for cx, cy in corner_points:
            if 0 <= cx < local_rect.width and 0 <= cy < local_rect.height:
                panel_overlay.set_at((cx, cy), (0, 0, 0, 0))
        self.image.blit(panel_overlay, panel_rect.topleft)

    def _paint_damage_marks(self: Self, *, health_ratio: float) -> None:
        visibility_scale = 0.7
        paint_wall_damage_overlay(
            self.image,
            health_ratio=health_ratio,
            seed=self._damage_visual_seed,
            variant_index=self._damage_overlay_variant,
            visibility_scale=visibility_scale,
        )
        if self.palette_category == "inner_wall" and 0.0 < health_ratio < 1.0:
            outline_color = resolve_wall_outline_color(
                palette_category=self.palette_category,
                palette=self.palette,
            )
            _draw_dotted_rect_outline(self.image, color=outline_color)

    def _apply_damage_alpha(self: Self, *, health_ratio: float) -> None:
        if self.palette_category != "inner_wall":
            self.image.set_alpha(255)
            return
        ratio = max(0.0, min(1.0, health_ratio))
        alpha = int(round(255 * (0.5 + 0.5 * ratio)))
        self.image.set_alpha(max(0, min(255, alpha)))

    def collides_rect(self: Self, rect_obj: rect.Rect) -> bool:
        if self._collision_polygon is None:
            return self.rect.colliderect(rect_obj)
        return _rect_polygon_collision(rect_obj, self._collision_polygon)

    def _collides_circle(
        self: Self, center: tuple[float, float], radius: float
    ) -> bool:
        if not _circle_rect_collision(center, radius, self.rect):
            return False
        if self._collision_polygon is None:
            return True
        return _circle_polygon_collision(center, radius, self._collision_polygon)

    def set_palette(
        self: Self, palette: EnvironmentPalette | None, *, force: bool = False
    ) -> None:
        """Update the wall's palette to match the current ambient palette."""

        if not force and self.palette is palette:
            return
        self.palette = palette
        self._update_color()


class RubbleWall(Wall):
    def __init__(
        self: Self,
        x: int,
        y: int,
        width: int,
        height: int,
        *,
        health: int = INTERNAL_WALL_HEALTH,
        palette: EnvironmentPalette | None = None,
        palette_category: str = "inner_wall",
        bevel_depth: int = INTERNAL_WALL_BEVEL_DEPTH,
        rubble_rotation_deg: float | None = None,
        rubble_offset_px: int | None = None,
        damage_overlay_seed: int = _WALL_DAMAGE_OVERLAY_SEED,
        on_destroy: Callable[[Self], None] | None = None,
    ) -> None:
        self._rubble_rotation_deg = (
            RUBBLE_ROTATION_DEG if rubble_rotation_deg is None else rubble_rotation_deg
        )
        base_size = max(1, min(width, height))
        self._rubble_offset_px = (
            rubble_offset_for_size(base_size)
            if rubble_offset_px is None
            else rubble_offset_px
        )
        super().__init__(
            x,
            y,
            width,
            height,
            health=health,
            palette=palette,
            palette_category=palette_category,
            bevel_depth=bevel_depth,
            bevel_mask=(False, False, False, False),
            draw_bottom_side=False,
            damage_overlay_seed=damage_overlay_seed,
            on_destroy=on_destroy,
        )

    def _update_color(self: Self) -> None:
        if self.health <= 0:
            health_ratio = 0.0
        else:
            health_ratio = max(0.0, self.health / self.max_health)
        fill_color, _border_color = resolve_wall_colors(
            health_ratio=health_ratio,
            palette_category=self.palette_category,
            palette=self.palette,
        )
        rubble_surface = build_rubble_wall_surface(
            self.image.get_width(),
            fill_color=fill_color,
            palette_category=self.palette_category,
            palette=self.palette,
            angle_deg=self._rubble_rotation_deg,
            offset_px=self._rubble_offset_px,
            bevel_depth=self.bevel_depth,
            relief_variant_index=self._damage_overlay_variant,
            draw_outline=not (0.0 < health_ratio < 1.0),
        )
        self.image.fill((0, 0, 0, 0))
        self.image.blit(rubble_surface, (0, 0))
        self._paint_damage_marks(health_ratio=health_ratio)
        self._apply_damage_alpha(health_ratio=health_ratio)


class ReinforcedWall(Wall):
    """Non-destructible-looking inner wall with metallic frame highlights."""

    def __init__(
        self: Self,
        x: int,
        y: int,
        width: int,
        height: int,
        *,
        health: int = INTERNAL_WALL_HEALTH,
        palette: EnvironmentPalette | None = None,
        bevel_depth: int = INTERNAL_WALL_BEVEL_DEPTH,
        bevel_mask: tuple[bool, bool, bool, bool] | None = None,
        draw_bottom_side: bool = False,
        bottom_side_ratio: float = 0.1,
        side_shade_ratio: float = 0.9,
        on_destroy: Callable[[Self], None] | None = None,
    ) -> None:
        super().__init__(
            x,
            y,
            width,
            height,
            health=health,
            palette=palette,
            palette_category="outer_wall",
            bevel_depth=bevel_depth,
            bevel_mask=bevel_mask,
            draw_bottom_side=draw_bottom_side,
            bottom_side_ratio=bottom_side_ratio,
            side_shade_ratio=side_shade_ratio,
            on_destroy=on_destroy,
        )

    def _update_color(self: Self) -> None:
        if self.health <= 0:
            health_ratio = 0.0
        else:
            health_ratio = max(0.0, self.health / self.max_health)
        fill_color, border_color = resolve_wall_colors(
            health_ratio=health_ratio,
            palette_category="outer_wall",
            palette=self.palette,
        )
        outline_color = resolve_wall_outline_color(
            palette_category="outer_wall",
            palette=self.palette,
        )
        paint_wall_surface(
            self.image,
            fill_color=fill_color,
            outline_color=outline_color,
            bevel_depth=self.bevel_depth,
            bevel_mask=self.bevel_mask,
            draw_bottom_side=self.draw_bottom_side,
            bottom_side_ratio=self.bottom_side_ratio,
            side_shade_ratio=self.side_shade_ratio,
            draw_outline=True,
        )

        w, h = self.image.get_size()
        side_height = (
            max(1, int(h * max(0.0, self.bottom_side_ratio)))
            if self.draw_bottom_side
            else 0
        )
        top_height = max(1, h - side_height)
        frame_width = max(2, min(w, top_height) // 6)
        inset = max(4, frame_width + 2)
        frame_rect = pygame.Rect(
            inset,
            inset,
            max(4, w - inset * 2),
            max(4, top_height - inset * 2),
        )

        frame_color = (
            max(0, int(border_color[0] * 0.656)),
            max(0, int(border_color[1] * 0.656)),
            max(0, int(border_color[2] * 0.704)),
        )
        highlight = (
            min(255, int(fill_color[0] * 1.14)),
            min(255, int(fill_color[1] * 1.14)),
            min(255, int(fill_color[2] * 1.14)),
        )
        thin_highlight = (
            min(255, int(fill_color[0] * 1.22)),
            min(255, int(fill_color[1] * 1.22)),
            min(255, int(fill_color[2] * 1.22)),
        )

        pygame.draw.rect(self.image, frame_color, frame_rect)

        frame_overlay = pygame.Surface((w, h), pygame.SRCALPHA)
        base_step = max(5, frame_width)
        rhythm = [
            base_step,
            max(3, base_step - 2),
            base_step + 2,
            max(4, base_step - 1),
        ]
        thick_width = max(1, frame_width // 3)
        sx = frame_rect.left - frame_rect.height
        i = 0
        while sx <= frame_rect.right + frame_rect.height:
            start = (sx, frame_rect.bottom)
            end = (sx + frame_rect.height, frame_rect.top)
            pygame.draw.line(frame_overlay, highlight, start, end, width=thick_width)
            thin_offset = max(2, thick_width + 1)
            pygame.draw.line(
                frame_overlay,
                thin_highlight,
                (start[0] + thin_offset, start[1]),
                (end[0] + thin_offset, end[1]),
                width=1,
            )
            sx += rhythm[i % len(rhythm)]
            i += 1
        frame_mask = pygame.Surface((w, h), pygame.SRCALPHA)
        pygame.draw.rect(frame_mask, (255, 255, 255, 255), frame_rect)
        pygame.draw.rect(
            frame_mask,
            (0, 0, 0, 0),
            pygame.Rect(0, top_height, w, h - top_height),
        )
        frame_overlay.blit(frame_mask, (0, 0), special_flags=pygame.BLEND_RGBA_MULT)
        self.image.blit(frame_overlay, (0, 0))

        inner_rect = frame_rect.inflate(-frame_width * 2, -frame_width * 2)
        if inner_rect.width > 0 and inner_rect.height > 0:
            panel_color = fill_color
            pygame.draw.rect(self.image, panel_color, inner_rect)
        self._paint_damage_marks(health_ratio=health_ratio)


class SteelBeam(pygame.sprite.Sprite):
    """Single-cell obstacle that behaves like a tougher internal wall."""

    def __init__(
        self: Self,
        x: int,
        y: int,
        size: int,
        *,
        health: int = STEEL_BEAM_HEALTH,
        palette: EnvironmentPalette | None = None,
        damage_overlay_seed: int = _WALL_DAMAGE_OVERLAY_SEED,
        on_destroy: Callable[[Self], None] | None = None,
    ) -> None:
        super().__init__()
        # Slightly inset from the cell size so it reads as a separate object.
        margin = max(3, size // 14)
        inset_size = max(4, size - margin * 2)
        self.image = pygame.Surface((inset_size, inset_size), pygame.SRCALPHA)
        self._added_to_groups = False
        self.health = health
        self.max_health = max(1, health)
        self.palette = palette
        self.on_destroy = on_destroy
        self._damage_visual_seed = damage_overlay_seed
        self._damage_overlay_variant = _damage_overlay_variant_index(
            x=x,
            y=y,
            width=size,
            height=size,
        )
        self._update_color()
        self.rect = self.image.get_rect(center=(x + size // 2, y + size // 2))

    def _take_damage(self: Self, *, amount: int = 1) -> None:
        if self.health > 0:
            self.health -= amount
            self._update_color()
            if self.health <= 0:
                if self.on_destroy is not None:
                    self.on_destroy(self)
                _mark_wall_index_dirty()
                self.kill()

    def _update_color(self: Self) -> None:
        """Render a simple square with crossed diagonals that darkens as damaged."""
        if self.health <= 0:
            return
        health_ratio = max(0.0, self.health / self.max_health)
        base_color, line_color = resolve_steel_beam_colors(
            health_ratio=health_ratio, palette=self.palette
        )
        paint_steel_beam_surface(
            self.image,
            base_color=base_color,
            line_color=line_color,
            health_ratio=health_ratio,
        )
        paint_wall_damage_overlay(
            self.image,
            health_ratio=health_ratio,
            seed=self._damage_visual_seed,
            variant_index=self._damage_overlay_variant,
        )


def _build_beveled_polygon(
    width: int,
    height: int,
    depth: int,
    bevels: tuple[bool, bool, bool, bool],
) -> list[tuple[int, int]]:
    return build_beveled_polygon(width, height, depth, bevels)


def _is_inner_wall(wall: pygame.sprite.Sprite) -> bool:
    if isinstance(wall, SteelBeam):
        return True
    if isinstance(wall, Wall):
        return wall.palette_category == "inner_wall"
    return False



from __future__ import annotations

from typing import TYPE_CHECKING

import pygame

try:
    from typing import Self
except ImportError:  # pragma: no cover - Python 3.10 fallback
    from typing_extensions import Self

from ..entities_constants import MATERIAL_SIZE

if TYPE_CHECKING:  # pragma: no cover - typing-only imports
    from .carrier_bot import CarrierBot


class Material(pygame.sprite.Sprite):
    """Passive carryable object for carrier bots."""

    def __init__(self: Self, x: float, y: float, *, size: int = MATERIAL_SIZE) -> None:
        super().__init__()
        safe_size = max(4, int(size))
        self.image = pygame.Surface((safe_size, safe_size), pygame.SRCALPHA)
        body_color = (56, 60, 66)
        frame_color = (24, 27, 32)
        wire_color = (104, 110, 120)
        inset = max(1, safe_size // 10)
        left = inset
        right = safe_size - 1 - inset
        top = inset
        bottom = safe_size - 1 - inset
        body_rect = pygame.Rect(left, top, right - left + 1, bottom - top + 1)
        # Draw as stacked metal plates: top face + visible front cross-section.
        front_depth = max(2, body_rect.height // 5)
        top_rect = pygame.Rect(
            body_rect.left,
            body_rect.top,
            body_rect.width,
            max(1, body_rect.height - front_depth),
        )
        front_rect = pygame.Rect(
            body_rect.left,
            top_rect.bottom,
            body_rect.width,
            max(1, body_rect.bottom - top_rect.bottom + 1),
        )
        pygame.draw.rect(self.image, body_color, top_rect)
        pygame.draw.rect(self.image, frame_color, top_rect, width=1)

        front_base_color = (46, 49, 55)
        pygame.draw.rect(self.image, front_base_color, front_rect)
        pygame.draw.rect(self.image, frame_color, front_rect, width=1)
        plate_line_colors = ((70, 74, 82), (58, 62, 70))
        plate_lines = max(2, min(4, front_rect.height))
        for i in range(1, plate_lines):
            y_pos = front_rect.top + (i * front_rect.height) // plate_lines
            pygame.draw.line(
                self.image,
                plate_line_colors[i % 2],
                (front_rect.left + 1, y_pos),
                (front_rect.right - 1, y_pos),
                width=1,
            )

        pygame.draw.rect(self.image, frame_color, body_rect, width=1)
        # Two vertical and two horizontal wire bands.
        # Vertical bands continue across the front cross-section.
        wire_width = max(1, safe_size // 10)
        x_positions = (safe_size // 3, (safe_size * 2) // 3)
        y_positions = (
            top_rect.top + max(1, top_rect.height // 3),
            top_rect.top + max(1, (top_rect.height * 2) // 3),
        )
        for x_pos in x_positions:
            pygame.draw.line(
                self.image,
                wire_color,
                (x_pos, top + 1),
                (x_pos, bottom - 1),
                width=wire_width,
            )
        for y_pos in y_positions:
            pygame.draw.line(
                self.image,
                wire_color,
                (left + 1, y_pos),
                (right - 1, y_pos),
                width=wire_width,
            )
        self.rect = self.image.get_rect(center=(int(x), int(y)))
        self.x = float(self.rect.centerx)
        self.y = float(self.rect.centery)
        self.collision_radius = float(max(1.0, safe_size * 0.5))
        # Entity-shadow system uses radius; 0.8x icon width => 0.4x radius.
        self.shadow_radius = max(1, int(safe_size * 0.4))
        self.shadow_shape = "rect"
        self.shadow_size = (
            max(1, int(safe_size * 1.2)),
            max(1, int(safe_size * 1.2)),
        )
        self.shadow_offset_scale = 1.0
        self.carried_by: CarrierBot | None = None

    def place_at(self: Self, x: float, y: float) -> None:
        self.x = float(x)
        self.y = float(y)
        self.rect.center = (int(self.x), int(self.y))



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

import pygame

try:
    from typing import Self
except ImportError:  # pragma: no cover - Python 3.10 fallback
    from typing_extensions import Self

from ..entities_constants import (
    CARRIER_BOT_COLLISION_RADIUS,
    CARRIER_BOT_HEIGHT,
    CARRIER_BOT_SPEED,
    CARRIER_BOT_WIDTH,
)
from ..render_constants import (
    PATROL_BOT_ARROW_COLOR,
    PATROL_BOT_BODY_COLOR,
    PATROL_BOT_OUTLINE_COLOR,
)
from .base_line_bot import BaseLineBot
from .material import Material
from .movement import _circle_wall_collision
from .walls import Wall

if TYPE_CHECKING:  # pragma: no cover - typing-only imports
    from ..models import LevelLayout


class CarrierBot(BaseLineBot):
    """Line-based carrier bot that can pick up and drop one material."""

    def __init__(
        self: Self,
        x: float,
        y: float,
        *,
        axis: str = "x",
        direction_sign: int = 1,
        speed: float = CARRIER_BOT_SPEED,
    ) -> None:
        super().__init__()
        self.axis = "y" if axis == "y" else "x"
        self.image = pygame.Surface((CARRIER_BOT_WIDTH, CARRIER_BOT_HEIGHT), pygame.SRCALPHA)
        self.rect = self.image.get_rect(center=(int(x), int(y)))
        self.x = float(self.rect.centerx)
        self.y = float(self.rect.centery)
        self.collision_radius = float(CARRIER_BOT_COLLISION_RADIUS)
        self.shadow_radius = max(1, int(self.collision_radius * 1.2))
        self.shadow_shape = "circle"
        self.shadow_offset_scale = 1.0
        self.speed = max(0.1, float(speed))
        self.direction = self._initial_direction(sign=direction_sign)
        self._refresh_surface()
        self.carried_material: Material | None = None
        self._recently_dropped_material: Material | None = None

    def _refresh_surface(self: Self) -> None:
        center = self.rect.center
        self.image = self._build_surface(axis=self.axis)
        self.rect = self.image.get_rect(center=center)

    def _reverse_direction(self: Self) -> None:
        super()._reverse_direction()
        self._refresh_surface()

    def _build_surface(self: Self, *, axis: str) -> pygame.Surface:
        surface = pygame.Surface((CARRIER_BOT_WIDTH, CARRIER_BOT_HEIGHT), pygame.SRCALPHA)
        rect = surface.get_rect()
        pygame.draw.rect(surface, PATROL_BOT_BODY_COLOR, rect)
        pygame.draw.rect(surface, PATROL_BOT_OUTLINE_COLOR, rect, width=2)
        # Slightly rounded corners: clear one pixel at each corner.
        surface.set_at((rect.left, rect.top), (0, 0, 0, 0))
        surface.set_at((rect.right - 1, rect.top), (0, 0, 0, 0))
        surface.set_at((rect.left, rect.bottom - 1), (0, 0, 0, 0))
        surface.set_at((rect.right - 1, rect.bottom - 1), (0, 0, 0, 0))

        cx = rect.width // 2
        cy = rect.height // 2
        marker_depth = max(2, min(rect.width, rect.height) // 6)
        marker_half = max(3, marker_depth * 2)
        offset = max(3, marker_depth + 2) + 1
        active_color = PATROL_BOT_ARROW_COLOR
        inactive_color = tuple(
            min(255, int((c * 0.35) + (255 * 0.65))) for c in PATROL_BOT_ARROW_COLOR
        )

        if axis == "y":
            up_tip_y = cy - offset
            up_base_y = up_tip_y + marker_depth
            up = [(cx, up_tip_y), (cx - marker_half, up_base_y), (cx + marker_half, up_base_y)]
            # Mirror the same integer-grid triangle to keep up/down shape identical.
            down = [(px, rect.height - 1 - py) for px, py in up]
            up_color = active_color if self.direction[1] < 0 else inactive_color
            down_color = active_color if self.direction[1] > 0 else inactive_color
            pygame.draw.polygon(surface, up_color, up)
            pygame.draw.polygon(surface, down_color, down)
        else:
            left_tip_x = cx - offset
            left_base_x = left_tip_x + marker_depth
            left = [(left_tip_x, cy), (left_base_x, cy - marker_half), (left_base_x, cy + marker_half)]
            # Mirror the same integer-grid triangle to keep left/right shape identical.
            right = [(rect.width - 1 - px, py) for px, py in left]
            left_color = active_color if self.direction[0] < 0 else inactive_color
            right_color = active_color if self.direction[0] > 0 else inactive_color
            pygame.draw.polygon(surface, left_color, left)
            pygame.draw.polygon(surface, right_color, right)
        return surface

    def _initial_direction(self: Self, *, sign: int) -> tuple[int, int]:
        step = 1 if sign >= 0 else -1
        if self.axis == "y":
            return (0, step)
        return (step, 0)

    def _sync_carried_material(self: Self) -> None:
        material = self.carried_material
        if material is None:
            return
        material.place_at(self.x, self.y)

    def _place_material_here(self: Self) -> None:
        material = self.carried_material
        if material is None:
            return
        material.carried_by = None
        material.place_at(self.x, self.y)
        self.carried_material = None
        self._recently_dropped_material = material

    def _cell_center(self: Self, *, cell: tuple[int, int], cell_size: int) -> tuple[float, float]:
        return (
            float((cell[0] * cell_size) + (cell_size / 2)),
            float((cell[1] * cell_size) + (cell_size / 2)),
        )

    def _can_drop_on_cell(
        self: Self,
        *,
        cell: tuple[int, int],
        cell_size: int,
        layout: "LevelLayout",
        pitfall_cells: set[tuple[int, int]],
        walls: list[Wall],
        materials: Iterable[Material],
        blockers: Iterable[pygame.sprite.Sprite],
    ) -> bool:
        cx, cy = cell
        if cell_size <= 0:
            return False
        if not (0 <= cx < layout.grid_cols and 0 <= cy < layout.grid_rows):
            return False
        if cell in pitfall_cells or cell in layout.outer_wall_cells or cell in layout.outside_cells:
            return False
        drop_x, drop_y = self._cell_center(cell=cell, cell_size=cell_size)
        carried = self.carried_material
        if carried is None:
            return False
        drop_radius = float(carried.collision_radius)
        for wall in walls:
            if _circle_wall_collision((drop_x, drop_y), drop_radius, wall):
                return False
        for material in materials:
            if material is carried or material.carried_by is not None:
                continue
            if int(material.rect.centerx // cell_size) == cx and int(
                material.rect.centery // cell_size
            ) == cy:
                return False
        for blocker in blockers:
            if blocker is self or blocker is carried:
                continue
            br = getattr(blocker, "collision_radius", None)
            if br is None:
                br = max(blocker.rect.width, blocker.rect.height) / 2
            dx = drop_x - float(blocker.rect.centerx)
            dy = drop_y - float(blocker.rect.centery)
            hit_range = drop_radius + float(br)
            if dx * dx + dy * dy <= hit_range * hit_range:
                return False
        return True

    def _place_material_on_grid(
        self: Self,
        *,
        cell_size: int,
        layout: "LevelLayout",
        pitfall_cells: set[tuple[int, int]],
        walls: list[Wall],
        materials: Iterable[Material],
        blockers: Iterable[pygame.sprite.Sprite],
    ) -> None:
        material = self.carried_material
        if material is None or cell_size <= 0:
            self._place_material_here()
            return
        base_cell = (int(self.x // cell_size), int(self.y // cell_size))
        forward = (int(self.direction[0]), int(self.direction[1]))
        backward = (-int(self.direction[0]), -int(self.direction[1]))
        forward_cell_1 = (base_cell[0] + forward[0], base_cell[1] + forward[1])
        forward_cell_2 = (base_cell[0] + forward[0] * 2, base_cell[1] + forward[1] * 2)
        backward_cell_1 = (base_cell[0] + backward[0], base_cell[1] + backward[1])
        backward_cell_2 = (base_cell[0] + backward[0] * 2, base_cell[1] + backward[1] * 2)
        target_forward_cells = {forward_cell_1, forward_cell_2}
        car_blocking_ahead = False
        for blocker in blockers:
            if blocker is self:
                continue
            if blocker.__class__.__name__ != "Car":
                continue
            blocker_cell = (
                int(blocker.rect.centerx // cell_size),
                int(blocker.rect.centery // cell_size),
            )
            if blocker_cell in target_forward_cells:
                car_blocking_ahead = True
                break
        if car_blocking_ahead:
            candidates = [
                forward_cell_1,
                forward_cell_2,
                base_cell,
                backward_cell_1,
                backward_cell_2,
            ]
        else:
            candidates = [
                base_cell,
                backward_cell_1,
                backward_cell_2,
                forward_cell_1,
                forward_cell_2,
            ]
        for cell in candidates:
            if not self._can_drop_on_cell(
                cell=cell,
                cell_size=cell_size,
                layout=layout,
                pitfall_cells=pitfall_cells,
                walls=walls,
                materials=materials,
                blockers=blockers,
            ):
                continue
            drop_x, drop_y = self._cell_center(cell=cell, cell_size=cell_size)
            material.carried_by = None
            material.place_at(drop_x, drop_y)
            self.carried_material = None
            self._recently_dropped_material = material
            return
        self._place_material_here()

    def _try_pick_overlapping_material(
        self: Self,
        *,
        materials: Iterable[Material],
    ) -> bool:
        for material in materials:
            if material.carried_by is not None:
                continue
            if material is self._recently_dropped_material:
                continue
            dx = self.x - float(material.rect.centerx)
            dy = self.y - float(material.rect.centery)
            # "Complete overlap": either circle is fully inside the other.
            pickup_range = abs(self.collision_radius - material.collision_radius)
            if dx * dx + dy * dy > pickup_range * pickup_range:
                continue
            material.carried_by = self
            self.carried_material = material
            self._sync_carried_material()
            return True
        return False

    def _is_blocked(
        self: Self,
        *,
        next_x: float,
        next_y: float,
        walls: list[Wall],
        layout: "LevelLayout",
        cell_size: int,
        pitfall_cells: set[tuple[int, int]],
        blockers: Iterable[pygame.sprite.Sprite],
    ) -> bool:
        _, _, hit_wall = self._handle_axis_collision(
            next_x=next_x,
            next_y=next_y,
            current_x=self.x,
            current_y=self.y,
            walls=walls,
            radius=self.collision_radius,
        )
        if hit_wall:
            return True

        if not (
            0 <= next_x < layout.field_rect.width and 0 <= next_y < layout.field_rect.height
        ):
            return True

        if cell_size > 0:
            lead_x = next_x + float(self.direction[0]) * (self.collision_radius + 1.0)
            lead_y = next_y + float(self.direction[1]) * (self.collision_radius + 1.0)
            lead_cell = (int(lead_x // cell_size), int(lead_y // cell_size))
            if lead_cell in pitfall_cells:
                return True
            if lead_cell in layout.outer_wall_cells or lead_cell in layout.outside_cells:
                return True

        for blocker in blockers:
            if blocker is self:
                continue
            if blocker is self.carried_material:
                continue
            if blocker is self._recently_dropped_material:
                continue
            br = getattr(blocker, "collision_radius", None)
            if br is None:
                br = max(blocker.rect.width, blocker.rect.height) / 2
            dx = next_x - float(blocker.rect.centerx)
            dy = next_y - float(blocker.rect.centery)
            hit_range = self.collision_radius + float(br)
            if dx * dx + dy * dy <= hit_range * hit_range:
                return True
        return False

    def update(
        self: Self,
        walls: list[Wall],
        *,
        layout: "LevelLayout",
        cell_size: int,
        pitfall_cells: set[tuple[int, int]],
        materials: Iterable[Material],
        blockers: Iterable[pygame.sprite.Sprite] = (),
        push_targets: Iterable[pygame.sprite.Sprite] | None = None,
    ) -> None:
        recent = self._recently_dropped_material
        if recent is not None:
            if recent.carried_by is not None:
                self._recently_dropped_material = None
            else:
                dx = self.x - float(recent.rect.centerx)
                dy = self.y - float(recent.rect.centery)
                release_dist = (
                    self.collision_radius + recent.collision_radius + max(1.0, self.speed)
                )
                if dx * dx + dy * dy >= release_dist * release_dist:
                    self._recently_dropped_material = None

        # Pick material only when physically overlapping it.
        if self.carried_material is None and self._try_pick_overlapping_material(
            materials=materials
        ):
            self._reverse_direction()
            return

        move_x = float(self.direction[0]) * self.speed
        move_y = float(self.direction[1]) * self.speed
        next_x = self.x + move_x
        next_y = self.y + move_y
        all_blockers = list(blockers)
        if self.carried_material is not None:
            all_blockers.extend([m for m in materials if m is not self.carried_material])

        if self._is_blocked(
            next_x=next_x,
            next_y=next_y,
            walls=walls,
            layout=layout,
            cell_size=cell_size,
            pitfall_cells=pitfall_cells,
            blockers=all_blockers,
        ):
            if self.carried_material is not None:
                self._place_material_on_grid(
                    cell_size=cell_size,
                    layout=layout,
                    pitfall_cells=pitfall_cells,
                    walls=walls,
                    materials=materials,
                    blockers=all_blockers,
                )
            self._push_overlapping_entities(
                center_x=self.x,
                center_y=self.y,
                radius=self.collision_radius,
                entities=blockers if push_targets is None else push_targets,
                ignore=[self.carried_material] if self.carried_material else (),
            )
            self._reverse_direction()
            return

        self.x = next_x
        self.y = next_y
        self.rect.center = (int(self.x), int(self.y))
        self._sync_carried_material()
        self._push_overlapping_entities(
            center_x=self.x,
            center_y=self.y,
            radius=self.collision_radius,
            entities=blockers if push_targets is None else push_targets,
            ignore=[self.carried_material] if self.carried_material else (),
        )
        if self.carried_material is None and self._try_pick_overlapping_material(
            materials=materials
        ):
            self._reverse_direction()



from __future__ import annotations

import math

import pygame

from ..colors import DARK_RED, ORANGE, YELLOW, EnvironmentPalette
from ..render_constants import ANGLE_BINS
from .common import ANGLE_STEP

_CAR_UPSCALE_FACTOR = 4

_CAR_COLOR_SCHEMES: dict[str, dict[str, tuple[int, int, int]]] = {
    "default": {
        "healthy": YELLOW,
        "damaged": ORANGE,
        "critical": DARK_RED,
    },
    "disabled": {
        "healthy": (185, 185, 185),
        "damaged": (150, 150, 150),
        "critical": (110, 110, 110),
    },
}


def resolve_car_color(
    *,
    health_ratio: float,
    appearance: str,
    palette: EnvironmentPalette | None = None,
) -> tuple[int, int, int]:
    _ = palette
    color_scheme = _CAR_COLOR_SCHEMES.get(appearance, _CAR_COLOR_SCHEMES["default"])
    color = color_scheme["healthy"]
    if health_ratio < 0.6:
        color = color_scheme["damaged"]
    if health_ratio < 0.3:
        color = color_scheme["critical"]
    return color


def build_car_surface(width: int, height: int) -> pygame.Surface:
    return pygame.Surface((width, height), pygame.SRCALPHA)


def paint_car_surface(
    surface: pygame.Surface,
    *,
    width: int,
    height: int,
    color: tuple[int, int, int],
) -> None:
    upscale = _CAR_UPSCALE_FACTOR
    if upscale > 1:
        up_width = width * upscale
        up_height = height * upscale
        up_surface = pygame.Surface((up_width, up_height), pygame.SRCALPHA)
        _paint_car_surface_base(
            up_surface, width=up_width, height=up_height, color=color
        )
        scaled = pygame.transform.smoothscale(up_surface, (width, height))
        surface.fill((0, 0, 0, 0))
        surface.blit(scaled, (0, 0))
        return
    _paint_car_surface_base(surface, width=width, height=height, color=color)


def _paint_car_surface_base(
    surface: pygame.Surface,
    *,
    width: int,
    height: int,
    color: tuple[int, int, int],
) -> None:
    surface.fill((0, 0, 0, 0))

    trim_color = tuple(int(c * 0.6) for c in color)
    body_color = tuple(min(255, int(c * 1.15)) for c in color)
    tail_light_color = (255, 80, 50)
    headlight_color = (200, 200, 200)

    base_width = 150.0
    base_height = 210.0
    scale_x = width / base_width
    scale_y = height / base_height

    def _rect(x: float, y: float, w: float, h: float) -> pygame.Rect:
        return pygame.Rect(
            int(round(x * scale_x)),
            int(round(y * scale_y)),
            max(1, int(round(w * scale_x))),
            max(1, int(round(h * scale_y))),
        )

    def _radius(value: float) -> int:
        return max(1, int(round(value * min(scale_x, scale_y))))

    body_top = _rect(0, 0, 150, 140)
    body_bottom = _rect(0, 70, 150, 140)
    rear_bed = _rect(16, 98, 118, 88)

    pygame.draw.rect(surface, trim_color, body_top, border_radius=_radius(50))
    pygame.draw.rect(surface, trim_color, body_bottom, border_radius=_radius(37))
    pygame.draw.rect(surface, body_color, rear_bed)

    tail_left = _rect(30, 190, 30, 20)
    tail_right = _rect(90, 190, 30, 20)
    pygame.draw.rect(surface, tail_light_color, tail_left)
    pygame.draw.rect(surface, tail_light_color, tail_right)

    headlight_left = _rect(15, 7, 40, 20)
    headlight_right = _rect(95, 7, 40, 20)
    pygame.draw.ellipse(surface, headlight_color, headlight_left)
    pygame.draw.ellipse(surface, headlight_color, headlight_right)


def build_car_directional_surfaces(
    base_surface: pygame.Surface, *, bins: int = ANGLE_BINS
) -> list[pygame.Surface]:
    surfaces: list[pygame.Surface] = []
    upscale = _CAR_UPSCALE_FACTOR
    if upscale > 1:
        src_size = base_surface.get_size()
        upscale_surface = pygame.transform.scale(
            base_surface,
            (src_size[0] * upscale, src_size[1] * upscale),
        )
    else:
        upscale_surface = base_surface
    for idx in range(bins):
        angle_rad = idx * ANGLE_STEP
        rotation_deg = -math.degrees(angle_rad) - 90
        rotated = pygame.transform.rotate(upscale_surface, rotation_deg)
        if upscale > 1:
            scaled = pygame.transform.smoothscale(
                rotated,
                (
                    max(1, rotated.get_width() // upscale),
                    max(1, rotated.get_height() // upscale),
                ),
            )
            surfaces.append(scaled)
        else:
            surfaces.append(rotated)
    return surfaces



from __future__ import annotations

from typing import Iterable

import pygame

try:
    from typing import Self
except ImportError:  # pragma: no cover - Python 3.10 fallback
    from typing_extensions import Self

from .movement import _circle_wall_collision
from .walls import Wall


class BaseLineBot(pygame.sprite.Sprite):
    """Shared line-bot mechanics used by patrol/carrier variants."""

    direction: tuple[int, int]

    def _movement_axis(self: Self) -> str:
        """Return the current movement axis from direction state."""
        dx, dy = self.direction
        if abs(dx) >= abs(dy):
            return "x"
        return "y"

    def _forward_cell(
        self: Self,
        *,
        x: float,
        y: float,
        cell_size: int,
        step_cells: int = 1,
    ) -> tuple[int, int] | None:
        """Return the grid cell ahead of (x, y) along current direction."""
        if cell_size <= 0:
            return None
        if step_cells < 1:
            step_cells = 1
        dx, dy = self.direction
        step_x = 0
        step_y = 0
        if abs(dx) >= abs(dy):
            if dx > 0:
                step_x = step_cells
            elif dx < 0:
                step_x = -step_cells
        else:
            if dy > 0:
                step_y = step_cells
            elif dy < 0:
                step_y = -step_cells
        return int(x // cell_size) + step_x, int(y // cell_size) + step_y

    def _reverse_direction(self: Self) -> None:
        self.direction = (-self.direction[0], -self.direction[1])

    def _handle_axis_collision(
        self: Self,
        *,
        next_x: float,
        next_y: float,
        current_x: float,
        current_y: float,
        walls: list[Wall],
        radius: float,
    ) -> tuple[float, float, bool]:
        final_x, final_y = next_x, next_y
        hit = False
        if next_x != current_x:
            for wall in walls:
                if _circle_wall_collision((next_x, current_y), radius, wall):
                    final_x = current_x
                    hit = True
                    break
        if next_y != current_y:
            for wall in walls:
                if _circle_wall_collision((final_x, next_y), radius, wall):
                    final_y = current_y
                    hit = True
                    break
        return final_x, final_y, hit

    def _resolve_circle_overlap(
        self: Self,
        cx: float,
        cy: float,
        radius: float,
        *,
        other_x: float,
        other_y: float,
        other_radius: float,
    ) -> tuple[float, float]:
        dx = cx - other_x
        dy = cy - other_y
        dist_sq = dx * dx + dy * dy
        min_dist = radius + other_radius
        if dist_sq <= 0:
            return cx + min_dist, cy
        dist = dist_sq**0.5
        if dist >= min_dist:
            return cx, cy
        push = min_dist - dist
        return cx + (dx / dist) * push, cy + (dy / dist) * push

    def _push_overlapping_entities(
        self: Self,
        *,
        center_x: float,
        center_y: float,
        radius: float,
        entities: Iterable[pygame.sprite.Sprite],
        ignore: Iterable[pygame.sprite.Sprite] = (),
    ) -> None:
        ignored = set(ignore)
        for entity in entities:
            if entity is self or entity in ignored:
                continue
            if getattr(entity, "mounted_vehicle", None) is not None:
                continue
            ex = float(entity.rect.centerx)
            ey = float(entity.rect.centery)
            er = float(
                getattr(
                    entity,
                    "collision_radius",
                    getattr(entity, "radius", max(entity.rect.width, entity.rect.height) / 2),
                )
            )
            new_x, new_y = self._resolve_circle_overlap(
                ex,
                ey,
                er,
                other_x=center_x,
                other_y=center_y,
                other_radius=radius,
            )
            if new_x == ex and new_y == ey:
                continue
            entity.rect.center = (int(new_x), int(new_y))
            if hasattr(entity, "x"):
                entity.x = float(entity.rect.centerx)
            if hasattr(entity, "y"):
                entity.y = float(entity.rect.centery)
